/*****************************************************************************\
 *                                                                           *
 * Xrtti.cpp                                                                 *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/

#include <Essential/Hashtable.h>
#include <Essential/String.h>
#include <Essential/Vector.h>
#define private public // cheat the Xrtti object constructors/destructors
#define protected public // cheat the Xrtti object constructors/destructors
#include "Xrtti/XrttiPrivate.h"

using namespace Essential;


/** **************************************************************************
 * Convenience typedefs
 ************************************************************************** **/

typedef ::XrttiPrivate::Context XP_Context;
typedef ::XrttiPrivate::Enumeration XP_Enumeration;
typedef ::XrttiPrivate::Field XP_Field;
typedef ::XrttiPrivate::Method XP_Method;
typedef ::XrttiPrivate::Type XP_Type;
typedef ::Xrtti::S32 X_S32;
typedef ::Xrtti::U32 X_U32;
typedef ::Xrtti::Argument X_Argument;
typedef ::Xrtti::Array X_Array;
typedef ::Xrtti::Base X_Base;
typedef ::Xrtti::Class X_Class;
typedef ::Xrtti::Constructor X_Constructor;
typedef ::Xrtti::ConstructorSignature X_ConstructorSignature;
typedef ::Xrtti::Destructor X_Destructor;
typedef ::Xrtti::DestructorSignature X_DestructorSignature;
typedef ::Xrtti::Context X_Context;
typedef ::Xrtti::Enumeration X_Enumeration;
typedef ::Xrtti::EnumerationValue X_EnumerationValue;
typedef ::Xrtti::Field X_Field;
typedef ::Xrtti::Generated X_Generated;
typedef ::Xrtti::Incomplete X_Incomplete;
typedef ::Xrtti::Method X_Method;
typedef ::Xrtti::MethodSignature X_MethodSignature;
typedef ::Xrtti::Namespace X_Namespace;
typedef ::Xrtti::Pointer X_Pointer;
typedef ::Xrtti::Structure X_Structure;
typedef ::Xrtti::Struct X_Struct;
typedef ::Xrtti::Type X_Type;
typedef ::Xrtti::TypeEnumeration X_TypeEnumeration;
typedef ::Xrtti::TypeFunction X_TypeFunction;
typedef ::Xrtti::TypeStructure X_TypeStructure;
typedef ::Xrtti::TypeUnion X_TypeUnion;
typedef ::Xrtti::Union X_Union;
typedef ::Xrtti::Value X_Value;


/** **************************************************************************
 * Static Data Structures
 ************************************************************************** **/

static Hashtable<String, X_Context *> htContextsG;
static Hashtable<String, X_Class *> htClassesG;
static Hashtable<String, X_Incomplete *> htIncompletesG;
static Hashtable<String, X_Namespace *> htNamespacesG;
static Hashtable<String, X_Struct *> htStructsG;
static Hashtable<String, X_Union *> htUnionsG;
static Hashtable<String, X_Structure *> htTypeInfosG;
static Vector<X_Context *> vContextsG;
static Vector<X_Class *> vClassesG;
static Vector<X_Incomplete *> vIncompletesG;
static Vector<X_Namespace *> vNamespacesG;
static Vector<X_Struct *> vStructsG;
static Vector<X_Union *> vUnionsG;
static Vector<X_Type *> vTypesG;
static Vector<X_Enumeration *> vEnumerationsG;


/** **************************************************************************
 * Static Method Declarations
 ************************************************************************** **/

static X_Context *GetContext(const X_Generated &generated, X_U32 index);
static X_Structure *GetStructure(const X_Generated &generated, X_U32 index);
static X_Class *GetClass(const X_Generated &generated, 
                         const XP_Context &private_context);
static X_Incomplete *GetIncomplete(const X_Generated &generated, 
                                   const XP_Context &private_context);
static X_Namespace *GetNamespace(const X_Generated &generated,
                                 const XP_Context &private_context);
static X_Struct *GetStruct(const X_Generated &generated, 
                           const XP_Context &private_context);
static X_Union *GetUnion(const X_Generated &generated,
                         const XP_Context &private_context);
static X_Type *GetType(const X_Generated &generated, X_U32 index);
static X_Enumeration *GetEnumeration(const X_Generated &generated,
                                     X_U32 index);
static void InitializeContext(X_Context &xrtti_context,
                              const X_Generated &generated,
                              const XP_Context &private_context);
static void InitializeStructure(X_Structure &xrtti_structure,
                                const X_Generated &generated,
                                const XP_Context &private_context);
static void DeinitializeStructure(X_Structure &xrtti_structure);
static void InitializeStruct(X_Struct &xrtti_struct,
                             const X_Generated &generated,
                             const XP_Context &private_context);
static void DeinitializeStruct(X_Struct &xrtti_struct);
static void InitializeField(X_Context &xrtti_context,
                            X_Field &xrtti_field,
                            const X_Generated &generated,
                            const XP_Field &private_field);
static void InitializeDestructorSignature
    (X_DestructorSignature &xrtti_destructor_signature,
     const X_Generated &generated, const XP_Method &private_method);
static void DeinitializeDestructorSignature
    (X_DestructorSignature &xrtti_destructor_signature);
static void InitializeDestructor(X_Structure &xrtti_structure,
                                 X_Destructor &xrtti_destructor,
                                 const X_Generated &generated,
                                 const XP_Method &private_method);
static void InitializeConstructorSignature
    (X_ConstructorSignature &xrtti_constructor_signature,
     const X_Generated &generated, const XP_Method &private_method);
static void DeinitializeConstructorSignature
    (X_ConstructorSignature &xrtti_constructor_signature);
static void InitializeConstructor(X_Structure &xrtti_structure,
                                  X_Constructor &xrtti_constructor,
                                  const X_Generated &generated,
                                  const XP_Method &private_method);
static void DeinitializeConstructor(X_Constructor &xrtti_constructor);
static void InitializeMethodSignature
    (X_MethodSignature &xrtti_function_signature,
     const X_Generated &generated, const XP_Method &private_method);
static void InitializeMethod(X_Structure &xrtti_structure,
                             X_Method &xrtti_method,
                             const X_Generated &generated,
                             const XP_Method &private_method);
static void DeinitializeMethod(X_Method &xrtti_method);
static void InitializeType(X_Type &xrtti_type, const X_Generated &generated,
                           const XP_Type &private_type);
static void DeinitializeType(X_Type &xrtti_type);
static void InitializeArgument(X_Argument &xrtti_argument,
                               const X_Generated &generated, 
                               X_U32 type_index, void *pDefault);
static void InitializeTypeFunction(X_TypeFunction &xrtti_type_function,
                                   const X_Generated &generated,
                                   const XP_Type &private_type);
static void InitializeEnumeration(X_Enumeration &xrtti_enumeration,
                                  const X_Generated &generated,
                                  const XP_Enumeration &private_enumeration);
static void DeinitializeEnumeration(X_Enumeration &xrtti_enumeration);
static void AddContext(const String &fullName, X_Context *pContext);
static void SetValue(X_Value &value, const X_Type *pType, void *pDefault);


/** **************************************************************************
 * Statics Implementation
 ************************************************************************** **/

static X_Context *GetContext(const X_Generated &generated, X_U32 index)
{
    const XP_Context &private_context = *(generated.contexts[index]);

    switch (private_context.type) {
    case X_Context::Type_Class:
        return GetClass(generated, private_context);
    case X_Context::Type_Incomplete:
        return GetIncomplete(generated, private_context);
    case X_Context::Type_Namespace:
        return GetNamespace(generated, private_context);
    case X_Context::Type_Struct:
        return GetStruct(generated, private_context);
    default: // X_Context::Type_Union:
        return GetUnion(generated, private_context);
    }
}


static X_Structure *GetStructure(const X_Generated &generated, X_U32 index)
{
    const XP_Context &private_context = *(generated.contexts[index]);

    switch (private_context.type) {
    case X_Context::Type_Class:
        return GetClass(generated, private_context);
    case X_Context::Type_Struct:
        return GetStruct(generated, private_context);
    default: // X_Context::Type_Union:
        return GetUnion(generated, private_context);
    }
}


static X_Class *GetClass(const X_Generated &generated, 
                         const XP_Context &private_context)
{
    // Get the Class by full name, if there was one already
    String fullName(private_context.full_name);

    X_Class *pClass;

    if (!htClassesG.Get(fullName, /* returns */ pClass)) {
        // Create a new one
        pClass = new X_Class;
        // And put it in the hashtable
        X_Class *junk;
        htClassesG.Put(fullName, pClass, /* returns */ junk);
        // And in the vector
        vClassesG.Append(pClass);
        // And the context hashtable and vector
        AddContext(fullName, pClass);
        // And finally, in the type_info hashtable if possible
        if (private_context.typeinfo) {
            String typeinfo_name = private_context.typeinfo->name();
            X_Structure *pStructure = (X_Structure *) pClass, *junk2;
            htTypeInfosG.Put(typeinfo_name, pStructure, /* returns */  junk2);
        }
        // And initialize it
        InitializeStruct(*pClass, generated, private_context);
    }

    return pClass;
}


static X_Incomplete *GetIncomplete(const X_Generated &generated, 
                                   const XP_Context &private_context)
{
    // Get the Incomplete by full name, if there was one already
    String fullName(private_context.full_name);

    X_Incomplete *pIncomplete;

    if (!htIncompletesG.Get(fullName, /* returns */ pIncomplete)) {
        // Create a new one
        pIncomplete = new X_Incomplete;
        // And put it in the hashtable
        X_Incomplete *junk;
        htIncompletesG.Put(fullName, pIncomplete, /* returns */ junk);
        // And in the vector
        vIncompletesG.Append(pIncomplete);
        // And the context hashtable and vector
        AddContext(fullName, pIncomplete);
        // And initialize it
        InitializeContext(*pIncomplete, generated, private_context);
    }

    return pIncomplete;
}


static X_Namespace *GetNamespace(const X_Generated &generated,
                                 const XP_Context &private_context)
{
    // Get the Namespace by full name, if there was one already
    String fullName(private_context.full_name);

    X_Namespace *pNamespace;

    if (!htNamespacesG.Get(fullName, /* returns */ pNamespace)) {
        // Create a new one
        pNamespace = new X_Namespace;
        // And put it in the hashtable
        X_Namespace *junk;
        htNamespacesG.Put(fullName, pNamespace, /* returns */ junk);
        // And in the vector
        vNamespacesG.Append(pNamespace);
        // And the context hashtable and vector
        AddContext(fullName, pNamespace);
        // And initialize it
        InitializeContext(*pNamespace, generated, private_context);
    }

    return pNamespace;
}


static X_Struct *GetStruct(const X_Generated &generated, 
                           const XP_Context &private_context)
{
    // Get the Struct by full name, if there was one already
    String fullName(private_context.full_name);

    X_Struct *pStruct;

    if (!htStructsG.Get(fullName, /* returns */ pStruct)) {
        // Create a new one
        pStruct = new X_Struct;
        // And put it in the hashtable
        X_Struct *junk;
        htStructsG.Put(fullName, pStruct, /* returns */ junk);
        // And in the Vectors
        vStructsG.Append(pStruct);
        // And the context hashtable and vector
        AddContext(fullName, pStruct);
        // And finally, in the type_info hashtable if possible
        if (private_context.typeinfo) {
            String typeinfo_name = private_context.typeinfo->name();
            X_Structure *pStructure = (X_Structure *) pStruct, *junk2;
            htTypeInfosG.Put(typeinfo_name, pStructure, /* returns */ junk2);
        }
        // And initialize it
        InitializeStruct(*pStruct, generated, private_context);
    }

    return pStruct;
}


static X_Union *GetUnion(const X_Generated &generated, 
                         const XP_Context &private_context)
{
    // Get the Union by full name, if there was one already
    String fullName(private_context.full_name);

    X_Union *pUnion;

    if (!htUnionsG.Get(fullName, /* returns */ pUnion)) {
        // Create a new one
        pUnion = new X_Union;
        // And put it in the hashtables
        X_Union *junk;
        htUnionsG.Put(fullName, pUnion, /* returns */ junk);
        // And in the Vectors
        vUnionsG.Append(pUnion);
        // And the context hashtable and vector
        AddContext(fullName, pUnion);
        // And finally, in the type_info hashtable if possible
        if (private_context.typeinfo) {
            String typeinfo_name = private_context.typeinfo->name();
            X_Structure *pStructure = (X_Structure *) pUnion, *junk2;
            htTypeInfosG.Put(typeinfo_name, pStructure, /* returns */ junk2);
        }
        // And initialize it
        InitializeStructure(*pUnion, generated, private_context);
    }

    return pUnion;
}


static X_Type *GetType(const X_Generated &generated, X_U32 index)
{
    X_Type *pType = vTypesG[index];
    
    if (pType == NULL) {
        const XP_Type &private_type = *(generated.types[index]);
        if (private_type.type == X_Type::BaseType_Enumeration) {
            X_TypeEnumeration *pTypeEnumeration = new X_TypeEnumeration;
            vTypesG[index] = pTypeEnumeration;
            InitializeType(*pTypeEnumeration, generated, private_type);
            pTypeEnumeration->pEnumerationM = 
                GetEnumeration(generated, private_type.enumeration);
            return pTypeEnumeration;
        }
        else if (private_type.type == X_Type::BaseType_Function) {
            X_TypeFunction *pTypeFunction = new X_TypeFunction;
            vTypesG[index] = pTypeFunction;
            InitializeTypeFunction(*pTypeFunction, generated, private_type);
            return pTypeFunction;
        }
        else if (private_type.type == X_Type::BaseType_Structure) {
            X_TypeStructure *pTypeStructure = new X_TypeStructure;
            vTypesG[index] = pTypeStructure;
            InitializeType(*pTypeStructure, generated, private_type);
            pTypeStructure->pStructureM = 
                GetStructure(generated, private_type.structure);
            return pTypeStructure;
        }
        else {
            pType = new X_Type;
            vTypesG[index] = pType;
            InitializeType(*pType, generated, private_type);
        }
    }
    
    return pType;
}


static X_Enumeration *GetEnumeration(const X_Generated &generated, X_U32 index)
{
    X_Enumeration *pEnumeration = vEnumerationsG[index];
    
    if (pEnumeration == NULL) {
        pEnumeration = new X_Enumeration;
        vEnumerationsG[index] = pEnumeration;
        InitializeEnumeration(*pEnumeration, generated,
                              *(generated.enumerations[index]));
    }

    return pEnumeration;
}


static void InitializeContext(X_Context &xrtti_context,
                              const X_Generated &generated,
                              const XP_Context &private_context)
{
    // Context::pNameM
    xrtti_context.pNameM = private_context.name;

    // Context::pFullNameM
    xrtti_context.pFullNameM = private_context.full_name;

    // Context::pParentM
    if (private_context.parent == 0xFFFFFFFF) {
        xrtti_context.pParentM = NULL;
    }
    else {
        xrtti_context.pParentM = GetContext(generated, private_context.parent);
    }
}


static void InitializeStructure(X_Structure &xrtti_structure,
                                const X_Generated &generated, 
                                const XP_Context &private_context)
{
    // Context
    InitializeContext(xrtti_structure, generated, private_context);

    // accessTypeM
    xrtti_structure.accessTypeM = private_context.access_type;

    // pTypeInfoM
    xrtti_structure.pTypeInfoM = private_context.typeinfo;

    // Structure::baseCountM
    xrtti_structure.baseCountM = private_context.supers_count;

    // Structure::pBasesM
    if (private_context.supers_count) {
        xrtti_structure.pBasesM = 
            new X_Base[private_context.supers_count];
        for (X_U32 i = 0; i < private_context.supers_count; i++) {
            xrtti_structure.pBasesM[i].pStructureM =
                GetStructure(generated, private_context.supers[i]);
            if (private_context.supers_offsets) {
                xrtti_structure.pBasesM[i].offsetM = 
                    private_context.supers_offsets[i];
            }
            else {
                xrtti_structure.pBasesM[i].offsetM = 0;
            }
        }
    }
    else {
        xrtti_structure.pBasesM = NULL;
    }

    // Structure::fieldCountM
    xrtti_structure.fieldCountM = private_context.fields_count;

    // Structure::pFieldsM
    if (private_context.fields_count) {
        xrtti_structure.pFieldsM = new X_Field[private_context.fields_count];
        for (X_U32 i = 0; i < private_context.fields_count; i++) {
            InitializeField(xrtti_structure, xrtti_structure.pFieldsM[i],
                            generated, private_context.fields[i]);
        }
    }
    else {
        xrtti_structure.pFieldsM = NULL;
    }

    // Structure::isAnonymousM
    xrtti_structure.isAnonymousM = private_context.is_anonymous;

    // Structure::constructorCountM
    xrtti_structure.constructorCountM = private_context.constructors_count;

    // Structure::pConstructorsM
    if (private_context.constructors_count) {
        xrtti_structure.pConstructorsM = 
            new X_Constructor[private_context.constructors_count];
        for (X_U32 i = 0; i < private_context.constructors_count; i++) {
            InitializeConstructor
                (xrtti_structure, xrtti_structure.pConstructorsM[i],
                 generated, private_context.constructors[i]);
        }
    }
    else {
        xrtti_structure.pConstructorsM = NULL;
    }

    // Structure::pDestructorM
    if (private_context.destructor) {
        xrtti_structure.pDestructorM = new X_Destructor;
        InitializeDestructor(xrtti_structure, *(xrtti_structure.pDestructorM),
                             generated, *(private_context.destructor));
    }
    else {
        xrtti_structure.pDestructorM = NULL;
    }

    // Structure::pCreateMethodM
    xrtti_structure.pCreateMethodM = private_context.create_one;
    
    // Structure::pCreateArrayMethodM
    xrtti_structure.pCreateArrayMethodM = private_context.create_array;

    // Structure::pDeleteMethodM
    xrtti_structure.pDeleteMethodM = private_context.delete_one;

    // Structure::pDeleteArrayMethodM
    xrtti_structure.pDeleteArrayMethodM = private_context.delete_array;
}


static void DeinitializeStructure(X_Structure &xrtti_structure)
{
    delete[] xrtti_structure.pBasesM;

    delete[] xrtti_structure.pFieldsM;
    
    if (xrtti_structure.pConstructorsM) {
        for (X_U32 i = 0; i < xrtti_structure.constructorCountM; i++) {
            DeinitializeConstructor(xrtti_structure.pConstructorsM[i]);
        }
        delete[] xrtti_structure.pConstructorsM;
    }

    if (xrtti_structure.pDestructorM) {
        DeinitializeDestructorSignature
            (xrtti_structure.pDestructorM->signatureM);
        delete xrtti_structure.pDestructorM;
    }
}


static void InitializeStruct(X_Struct &xrtti_struct,
                             const X_Generated &generated,
                             const XP_Context &private_context)
{
    // Structure
    InitializeStructure(xrtti_struct, generated, private_context);

    // Struct::isAbstractM
    xrtti_struct.isAbstractM = false;
    for (U32 i = 0; i < private_context.methods_count; i++) {
        const XP_Method &xpm = private_context.methods[i];

        if (xpm.is_pure_virtual) {
            xrtti_struct.isAbstractM = true;
            break;
        }
    }

    // Struct::methodCountM
    xrtti_struct.methodCountM = private_context.methods_count;

    // Struct::pMethodsM
    if (private_context.methods_count) {
        xrtti_struct.pMethodsM = new X_Method[private_context.methods_count];
        for (X_U32 i = 0; i < private_context.methods_count; i++) {
            InitializeMethod(xrtti_struct, xrtti_struct.pMethodsM[i],
                             generated, private_context.methods[i]);
        }
    }
    else {
        xrtti_struct.pMethodsM = NULL;
    }
}


static void DeinitializeStruct(X_Struct &xrtti_struct)
{
    if (xrtti_struct.pMethodsM) {
        for (X_U32 i = 0; i < xrtti_struct.methodCountM; i++) {
            DeinitializeMethod(xrtti_struct.pMethodsM[i]);
        }
        delete[] xrtti_struct.pMethodsM;
    }

    DeinitializeStructure(xrtti_struct);
}


static void InitializeField(X_Context &xrtti_context, X_Field &xrtti_field,
                            const X_Generated &generated,
                            const XP_Field &private_field)
{
    // Member::accessTypeM
    xrtti_field.accessTypeM = private_field.access_type;

    // Member::pContextM
    xrtti_field.pContextM = &xrtti_context;

    // Member::pNameM
    xrtti_field.pNameM = private_field.name;
    
    // Member::isStaticM
    xrtti_field.isStaticM = private_field.is_static ? true : false;

    // Field::pTypeM
    xrtti_field.pTypeM = GetType(generated, private_field.type_index);

    // Field::offsetM
    xrtti_field.offsetM = private_field.offset;

    // Field::pGetMethodM
    xrtti_field.pGetMethodM = private_field.get;

    // Field::pSetMethodM
    xrtti_field.pSetMethodM = private_field.set;
}


static void InitializeDestructorSignature
    (X_DestructorSignature &xrtti_destructor_signature,
     const X_Generated &generated, const XP_Method &private_method)
{
    // Destructor::throwCountM
    xrtti_destructor_signature.throwCountM = private_method.throws_count;

    // Destructor::pThrowsM
    if (private_method.throws_count) {
        xrtti_destructor_signature.pThrowsM = 
            new X_Type *[private_method.throws_count];

        for (U32 i = 0; i < private_method.throws_count; i++) {
            xrtti_destructor_signature.pThrowsM[i] = 
                GetType(generated, private_method.throws[i]);
        }
    }
    else {
        xrtti_destructor_signature.pThrowsM = NULL;
    }
}


static void DeinitializeDestructorSignature
    (X_DestructorSignature &xrtti_destructor_signature)
{
    delete [] xrtti_destructor_signature.pThrowsM;
}


static void InitializeDestructor(X_Structure &xrtti_structure,
                                 X_Destructor &xrtti_destructor,
                                 const X_Generated &generated,
                                 const XP_Method &private_method)
{
    // Member::accessTypeM
    xrtti_destructor.accessTypeM = private_method.access_type;

    // Member::pContextM
    xrtti_destructor.pContextM = &xrtti_structure;

    // Destructor::isVirtualM
    xrtti_destructor.isVirtualM = private_method.is_virtual;

    // Destructor::isPureVirtualM
    xrtti_destructor.isPureVirtualM = private_method.is_pure_virtual;

    // Destructor::signatureM
    InitializeDestructorSignature(xrtti_destructor.signatureM, generated,
                                  private_method);
    
    // Destructor::pInvokeDestructorM
    xrtti_destructor.pInvokeDestructorM = xrtti_structure.pDeleteMethodM;
}


static void InitializeConstructorSignature
    (X_ConstructorSignature &xrtti_constructor_signature,
     const X_Generated &generated, const XP_Method &private_method)
{
    // DestructorSignature
    InitializeDestructorSignature
        ((X_DestructorSignature &) xrtti_constructor_signature,
         generated, private_method);

    // ConstructorSignature::argumentCountM
    xrtti_constructor_signature.argumentCountM = 
        private_method.arguments_count;

    // ConstructorSignature::pArgumentsM
    if (private_method.arguments_count) {
        xrtti_constructor_signature.pArgumentsM = 
            new X_Argument[private_method.arguments_count];
        for (X_U32 i = 0; i < private_method.arguments_count; i++) {
            InitializeArgument(xrtti_constructor_signature.pArgumentsM[i],
                               generated,
                               private_method.argument_types[i],
                               private_method.argument_defaults[i]);
        }
    }
    else {
        xrtti_constructor_signature.pArgumentsM = NULL;
    }

    // ConstructorSignature::hasEllipsisM
    xrtti_constructor_signature.hasEllipsisM = private_method.has_ellipsis;
}


static void DeinitializeConstructorSignature
    (X_ConstructorSignature &xrtti_constructor_signature)
{
    DeinitializeDestructorSignature(xrtti_constructor_signature);

    delete [] xrtti_constructor_signature.pArgumentsM;
}


static void InitializeConstructor(X_Structure &xrtti_structure,
                                  X_Constructor &xrtti_constructor,
                                  const X_Generated &generated,
                                  const XP_Method &private_method)
{
    // Member::accessTypeM
    xrtti_constructor.accessTypeM = private_method.access_type;

    // Member::pContextM
    xrtti_constructor.pContextM = &xrtti_structure;

    // Constructor::signatureM
    InitializeConstructorSignature(xrtti_constructor.signatureM,
                                   generated, private_method);

    // Constructor::pArgumentNamesM
    if (private_method.arguments_count) {
        xrtti_constructor.pArgumentNamesM = 
            new const char *[private_method.arguments_count];
        for (U32 i = 0; i < private_method.arguments_count; i++) {
            xrtti_constructor.pArgumentNamesM[i] = 
                private_method.argument_names[i];
        }
    }
    else {
        xrtti_constructor.pArgumentNamesM = NULL;
    }

    // Constructor::pInvokeConstructorM
    xrtti_constructor.pInvokeConstructorM = private_method.invoke;
}


static void DeinitializeConstructor(X_Constructor &xrtti_constructor)
{
    DeinitializeConstructorSignature(xrtti_constructor.signatureM);

    delete [] xrtti_constructor.pArgumentNamesM;
}


static void InitializeMethodSignature
    (X_MethodSignature &xrtti_function_signature,
     const X_Generated &generated, const XP_Method &private_method)
{
    // ConstructorSignature
    InitializeConstructorSignature
        (xrtti_function_signature, generated, private_method);

    // Method::pReturnTypeM
    xrtti_function_signature.pReturnTypeM = 
        GetType(generated, private_method.return_type_index);
}
     

static void InitializeMethod(X_Structure &xrtti_structure,
                             X_Method &xrtti_method,
                             const X_Generated &generated,
                             const XP_Method &private_method)
{
    // Member::accessTypeM
    xrtti_method.accessTypeM = private_method.access_type;

    // Member::pContextM
    xrtti_method.pContextM = &xrtti_structure;

    // Member::pNameM
    xrtti_method.pNameM = private_method.name;

    // Member::isStaticM
    xrtti_method.isStaticM = private_method.is_static ? true : false;

    // Method::isConstM
    xrtti_method.isConstM = private_method.is_const;

    // Method::isVirtualM
    xrtti_method.isVirtualM = private_method.is_virtual;

    // Method::isPureVirtualM
    xrtti_method.isPureVirtualM = private_method.is_pure_virtual;

    // Method::signatureM
    InitializeMethodSignature(xrtti_method.signatureM, generated,
                                private_method);
    
    // Method::pArgumentNamesM
    if (private_method.arguments_count) {
        xrtti_method.pArgumentNamesM = 
            new const char *[private_method.arguments_count];
        for (U32 i = 0; i < private_method.arguments_count; i++) {
            xrtti_method.pArgumentNamesM[i] = private_method.argument_names[i];
        }
    }
    else {
        xrtti_method.pArgumentNamesM = NULL;
    }

    // Method::pInvokeMethodM
    xrtti_method.pInvokeMethodM = private_method.invoke;
}


static void DeinitializeMethod(X_Method &xrtti_method)
{
    DeinitializeConstructorSignature(xrtti_method.signatureM);
    
    delete[] xrtti_method.pArgumentNamesM;
}


static void InitializeType(X_Type &xrtti_type, const X_Generated &generated,
                           const XP_Type &private_type)
{
    // Type::baseTypeM
    xrtti_type.baseTypeM = private_type.type;

    // Type::isConstM
    xrtti_type.isConstM = private_type.is_const;
    
    // Type::isVolatileM
    xrtti_type.isVolatileM = private_type.is_volatile;

    // Type::isReferenceM
    xrtti_type.isReferenceM = private_type.is_reference;

    // Type::bitCountM
    xrtti_type.bitCountM = private_type.bit_count;

    // Type::pointerCountM
    xrtti_type.pointerCountM = private_type.pointers_count;

    // Type::pPointersM
    if (private_type.pointers_count) {
        xrtti_type.pPointersM = new X_Pointer[private_type.pointers_count];
        for (X_U32 i = 0; i < private_type.pointers_count; i++) {
            X_Pointer &pointer = xrtti_type.pPointersM[i];
            pointer.isConstM = private_type.pointers_const[i];
            pointer.isVolatileM = private_type.pointers_volatile[i];
        }
    }
    else {
        xrtti_type.pPointersM = NULL;
    }

    // Type::arrayCountM
    xrtti_type.arrayCountM = private_type.arrays_count;

    // Type::pArraysM
    if (private_type.arrays_count) {
        xrtti_type.pArraysM = new X_Array[private_type.arrays_count];
        for (X_U32 i = 0; i < private_type.arrays_count; i++) {
            X_Array &array = xrtti_type.pArraysM[i];
            array.isUnboundedM = private_type.arrays_unbounded[i];
            array.elementCountM = private_type.arrays_element_counts[i];
        }
    }
    else {
        xrtti_type.pArraysM = NULL;
    }
}


static void DeinitializeType(X_Type &xrtti_type)
{
    if (xrtti_type.baseTypeM == X_Type::BaseType_Function) {
        DeinitializeConstructorSignature
            (((X_TypeFunction *) &xrtti_type)->signatureM);
    }

    delete[] xrtti_type.pPointersM;

    delete[] xrtti_type.pArraysM;
}


static void InitializeTypeFunction(X_TypeFunction &xrtti_type_function,
                                   const X_Generated &generated,
                                   const XP_Type &private_type)
{
    // Type
    InitializeType(xrtti_type_function, generated, private_type);

    // TypeFunction::signatureM
    InitializeMethodSignature(xrtti_type_function.signatureM,
                              generated, *(private_type.method));
}


static void InitializeEnumeration(X_Enumeration &xrtti_enumeration,
                                  const X_Generated &generated,
                                  const XP_Enumeration &private_enumeration)
{
    // Enumeration::accessTypeM
    xrtti_enumeration.accessTypeM = private_enumeration.access_type;

    // Enumeration::pContextM
    if (private_enumeration.context == 0xFFFFFFFF) {
        xrtti_enumeration.pContextM = NULL;
    }
    else {
        xrtti_enumeration.pContextM =
            GetContext(generated, private_enumeration.context);
    }

    // Enumeration::pNameM
    xrtti_enumeration.pNameM = private_enumeration.name;

    // Enumeration::valueCountM
    xrtti_enumeration.valueCountM = private_enumeration.values_count;

    // Enumeration::pValuesM
    if (private_enumeration.values_count) {
        xrtti_enumeration.pValuesM =
            new X_EnumerationValue[private_enumeration.values_count];
        for (X_U32 i = 0; i < private_enumeration.values_count; i++) {
            X_EnumerationValue &value = xrtti_enumeration.pValuesM[i];
            value.pNameM = private_enumeration.value_names[i];
            value.valueM = private_enumeration.value_values[i];
        }
    }
    else {
        xrtti_enumeration.pValuesM = NULL;
    }
}


static void DeinitializeEnumeration(X_Enumeration &xrtti_enumeration)
{
    delete[] xrtti_enumeration.pValuesM;
}


static void InitializeArgument(X_Argument &xrtti_argument, 
                               const X_Generated &generated,
                               X_U32 type_index, void *pDefault)
{
    // Argument::pTypeM
    xrtti_argument.pTypeM = GetType(generated, type_index);

    // Argument::hasDefaultM
    xrtti_argument.hasDefaultM = (pDefault == NULL) ? false : true;

    // Argument::defaultM
    if (xrtti_argument.hasDefaultM) {
        // Set the Value's union member depending upon the type
        SetValue(xrtti_argument.defaultM, xrtti_argument.pTypeM, pDefault);
    }
}


static void AddContext(const String &fullName, X_Context *pContext)
{
    X_Context *junk;

    htContextsG.Put(fullName, pContext, /* returns */ junk);
    vContextsG.Append(pContext);
}


static void SetValue(X_Value &value, const X_Type *pType, void *pDefault)
{
    if (pType->GetPointerCount() || pType->GetArrayCount()) {
        value.pointer_v = * (void **) pDefault;
        return;
    }

    switch (pType->GetBaseType()) {
    case X_Type::BaseType_Bool:
        value.bool_v = * (bool *) pDefault;
        break;
    case X_Type::BaseType_Char:
        value.char_v = * (signed char *) pDefault;
        break;
    case X_Type::BaseType_Unsigned_Char:
        value.unsigned_char_v = * (unsigned char *) pDefault;
        break;
    case X_Type::BaseType_WChar:
        value.wchar_v = * (wchar_t *) pDefault;
        break;
    case X_Type::BaseType_Short:
        value.short_v = * (signed short *) pDefault;
        break;
    case X_Type::BaseType_Unsigned_Short:
        value.unsigned_short_v = * (unsigned short *) pDefault;
        break;
    case X_Type::BaseType_Enumeration:
    case X_Type::BaseType_Int:
        value.int_v = * (signed int *) pDefault;
        break;
    case X_Type::BaseType_Unsigned_Int:
        value.unsigned_int_v = * (unsigned int *) pDefault;
        break;
    case X_Type::BaseType_Long:
        value.long_v = * (signed long *) pDefault;
        break;
    case X_Type::BaseType_Unsigned_Long:
        value.unsigned_long_v = * (unsigned long *) pDefault;
        break;
    case X_Type::BaseType_Long_Long:
        value.long_long_v = * (signed long long *) pDefault;
        break;
    case X_Type::BaseType_Unsigned_Long_Long:
        value.unsigned_long_long_v = * (unsigned long long *) pDefault;
        break;
    case X_Type::BaseType_Float:
        value.float_v = * (float *) pDefault;
        break;
    case X_Type::BaseType_Double:
        value.double_v = * (double *) pDefault;
        break;
    case X_Type::BaseType_Long_Double:
        value.long_double_v = * (long double *) pDefault;
        break;
    case X_Type::BaseType_Bitfield:
    case X_Type::BaseType_Function:
    case X_Type::BaseType_Void: // not possible
    default: // BaseType_Structure:
        value.pointer_v = * (void **) pDefault;
        break;
    }
}


/** **************************************************************************
 * Implementation
 ************************************************************************** **/

const X_Base &X_Structure::GetBase(U32 index) const
{
    return pBasesM[index];
}


const X_Field &X_Structure::GetField(U32 index) const
{
    return pFieldsM[index];
}


const X_Constructor &X_Structure::GetConstructor(U32 index) const
{
    return pConstructorsM[index];
}


const X_Destructor &X_Structure::GetDestructor() const
{
    return *pDestructorM;
}


const X_Method &X_Struct::GetMethod(U32 index) const
{
    return pMethodsM[index];
}


const X_Pointer &X_Type::GetPointer(U32 depth) const
{
    return pPointersM[depth];
}


const X_Array &X_Type::GetArray(U32 dimension) const
{
    return pArraysM[dimension];
}


const X_EnumerationValue &X_Enumeration::GetValue(U32 index) const
{
    return pValuesM[index];
}


void Xrtti::Register(const X_Generated *generated)
{
    for (X_U32 i = 0; i < generated->contexts_count; i++) {
        (void) ::GetContext(*generated, i);
    }
}


X_U32 Xrtti::GetContextCount()
{
    return vContextsG.Count();
}


const X_Context *Xrtti::GetContext(X_U32 index)
{
    if (index < vContextsG.Count()) {
        return vContextsG[index];
    }
    else {
        return NULL;
    }
}


const X_Context *Xrtti::LookupContext(const char *pFullName)
{
    X_Context *pRet = NULL;

    if (!htContextsG.Get(pFullName, /* returns */ pRet)) {
        return NULL;
    }
    
    return pRet;
}


const X_Structure *Xrtti::LookupStructure(const std::type_info &typeinfo)
{
    X_Structure *pRet = NULL;

    const char *name = typeinfo.name();

    if (!htTypeInfosG.Get(name, /* returns */ pRet)) {
        return NULL;
    }

    return pRet;
}


void Xrtti::ShutDown()
{
    // Classes
    U32 count = vClassesG.Count();
    for (X_U32 i = 0; i < count; i++) {
        DeinitializeStruct(*(vClassesG[i]));
        delete vClassesG[i];
    }

    // Incompletes
    count = vIncompletesG.Count();
    for (X_U32 i = 0; i < count; i++) {
        delete vIncompletesG[i];
    }

    // Namespaces
    count = vNamespacesG.Count();
    for (X_U32 i = 0; i < count; i++) {
        // DeinitializeContext(*(vNamespacesG[i]));
        delete vNamespacesG[i];
    }

    // Structs
    count = vStructsG.Count();
    for (X_U32 i = 0; i < count; i++) {
        DeinitializeStruct(*(vStructsG[i]));
        delete vStructsG[i];
    }

    // Unions
    count = vUnionsG.Count();
    for (X_U32 i = 0; i < count; i++) {
        DeinitializeStructure(*(vUnionsG[i]));
        delete vUnionsG[i];
    }

    // Types
    count = vTypesG.Count();
    for (X_U32 i = 0; i < count; i++) {
        DeinitializeType(*(vTypesG[i]));
        delete vTypesG[i];
    }

    // Enumerations
    count = vEnumerationsG.Count();
    for (X_U32 i = 0; i < count; i++) {
        DeinitializeEnumeration(*(vEnumerationsG[i]));
        delete vEnumerationsG[i];
    }
}
