/*****************************************************************************\
 *                                                                           *
 * xrttigen.cpp                                                              *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/

#include "../../inc/private/Generator.h"
#include "../../inc/private/Parser.h"

using namespace Essential;


// global configuration variables
bool g_warnings, g_verbose;

#if 0
static void PrintElement(const Parser::Element *pElement, const String &indent)
{
    Log::MessageF(indent + "Element: %S",
                  (const unichar *) pElement->GetElementName());
    
    String id = pElement->GetAttributeValue("id");

    if (id != "") {
        Log::MessageF(", id = %S\n", (const unichar *) id);
    }
    else {
        Log::Message("\n");
    }

    Parser::Element::AttributeEnumeration attributes(*pElement);

    while (attributes.HasElements()) {
        ref<Parser::Attribute> pAttribute = attributes.GetNextElement();

        Log::MessageF(indent + "    Attribute: %S = %S\n",
                      (const unichar *) pAttribute->GetName(),
                      (const unichar *) pAttribute->GetValue());
    }
        
    Parser::Element::SubElementEnumeration subElements(*pElement);
    
    while (subElements.HasElements()) {
        ref<Parser::Element> pSubElement = subElements.GetNextElement();
        
        PrintElement(pSubElement, indent + "    ");
    }
}
#endif

int main(int argc, char **argv)
{
    // Create a configuration from the program variables
    Configuration config(argc, argv);

    // This is the object which holds all of the raw data parsed out of
    // the gccxml input
    Parser parser;

    // Parse standard input
    parser.Parse(config.GetInputStream());

    // This is the generator which will generate output from the input
    // parsed elements
    Generator generator(config, parser);

    // Generate the output
    generator.Generate();

#if 0
    Parser::ElementEnumeration elements(parser);
    String indent;

    while (elements.HasElements()) {
        ref<Parser::Element> pElement = elements.GetNextElement();

        PrintElement(pElement, "");
    }
#endif
}
